package applications.collages;

import java.awt.*;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseEvent;
import java.util.*;
import terms.*;
import algebras.*;
import displays.*;
import util.*;
import gui.*;
import applications.*;

/** A display for objects belonging to the class <code>collage</code>.
  * Collages are displayed on an object of type <code>canvas</code>,
  * which allows dragging and zooming by means of the mouse. Initially,
  * the origin is placed at the center of the canvas.
  * @see canvas
  */
public class collageDisplay extends PSDisplay {

  private collage displayed = null;
  private displayCanvas theCanvas = new displayCanvas();
  private static double initialScaling = 10;
  private boolean autoZoom = false;
  private boolean clip = false;
  private static double autoZoomSize = .9;
  private static double zoomFactor = 1.5;
  private static String zoomIn = "zoom in";
  private static String zoomOut = "zoom out";
  private static String auto = "auto zoom";
  private static String manual = "manual zoom";
  private static String reset = "reset";
  private static String clipOn = "clipping on";
  private static String clipOff = "clipping off";
  private static String[][] manualCommands = {{zoomIn, zoomOut}, {auto, reset}};
  private static String[][] autoCommands = {{manual}};
  private static String[] clipOffCommands = {clipOn};
  private static String[] clipOnCommands = {clipOff};
  
  public collageDisplay() { 
    theCanvas.setSize(500,500);
    theCanvas.setScaling(initialScaling);
  }


/** This type of display can deal only with subclasses of
  * <code>collageAlgebra</code>; other algebras will be rejected.
  * @see collageAlgebra
  */
  public boolean acceptable(algebra a) {
    return a instanceof collageAlgebra;
  }

/** Collages are displayed in the way described above. */
  protected void displayObject(Object obj) {
    if (obj instanceof collage) displayed = (collage)obj;
    else displayed = null;
    theCanvas.invalidate();
    theCanvas.repaint();
  }

/** Returns the <code>canvas</code> on which the collages will be shown.
  * @see canvas
  */
  public Component visualizer() {
    return theCanvas;
  }

/** Available commands are "zoom in", "zoom out", "auto zoom / manual zoom", "reset",
    and "clipping on/off". */
  public list commands() {
    list result = new list();
    String [][] commands = autoZoom ? autoCommands : manualCommands;
    for (int i = 0; i < commands.length; i++) result.append(commands[i]);
    if (super.PSOutputPossible()) {
      if (clip) result.append(clipOnCommands); else result.append(clipOffCommands);
    }
    result.concat(super.commands());
    return result;
  }
  
  public void execute(String command) {
    if (zoomIn.equals(command)) {
      theCanvas.zoom(zoomFactor);
      theCanvas.invalidate();
    }
    else if (zoomOut.equals(command)) {
      theCanvas.zoom(1/zoomFactor);
      theCanvas.invalidate();
    }
    else if (reset.equals(command)) {
      theCanvas.enableScaling();
      theCanvas.setTranslation(0,0);
      theCanvas.setScaling(initialScaling);
      theCanvas.invalidate();
    }
    else if (auto.equals(command) || manual.equals(command)) {
      autoZoom = auto.equals(command);
      if (autoZoom) {
        theCanvas.disableScaling();
        theCanvas.invalidate();
      }
      else theCanvas.enableScaling();
    } 
    else if (clipOn.equals(command) || clipOff.equals(command)) {
      clip = clipOn.equals(command);
    }
    else {
      Cursor wait = Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR);
      theCanvas.setCursor(wait);
      super.execute(command);
    }
    theCanvas.repaint();
    Cursor normal = Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);
    theCanvas.setCursor(normal);
  }
    
  private void PSCoords(point p) {
    PSCoords(p.x, p.y);
  }
  
  protected final boolean PSOutputPossible() {
    return super.PSOutputPossible() && displayed != null;
  }
  
  protected final void PSOutput() {
    Enumeration parts = displayed.parts();
    while (parts.hasMoreElements()) {
      part p = (part)parts.nextElement();
      if (clip && !p.shown) continue;
      Color c = p.getColour();
      PSSetColour(((double)c.getRed())/255,
                           ((double)c.getGreen())/255,
                           ((double)c.getBlue())/255);
      double thickness = p.getThickness();
      if (p instanceof dot) PSOutputDot((dot)p, thickness); 
      else if (p instanceof polyPart) PSOutputPolyPart((polyPart)p, thickness);
      else if (p instanceof bezierPart) PSOutputBezierPart((bezierPart)p, thickness);
    }
    PSOutputPreamble();
  }
  
  private void PSOutputDot(dot p, double thickness) {
    point pos = p.getPosition();
    if (thickness < 0) {
      PSStartPath(pos.x, pos.y, pos.x, pos.y);
      PSNextPointOfPath(pos.x, pos.y);
      PSStrokePath(true);
    }
    else {
      PSStartPath(pos.x - thickness / 2, pos.y - thickness / 2,
        pos.x + thickness / 2, pos.y + thickness / 2);
      PSNextPointOfPath(pos.x, pos.y);
      PSCoords(pos.x, pos.y); PSWrite(" ");
      PSCoord(thickness / 2);
      PSWrite(" 0 360 arc\n");
      PSFillPath();
    }
  }
  
  private void PSOutputPolyPart(polyPart p, double thickness) {
    point[] points = p.getPoints();
    double[] bounds = p.bounds();
    PSStartPath(bounds[0], bounds[1], bounds[2], bounds[3]);
    for (int i = 0; i < points.length; i++) PSNextPointOfPath(points[i].x, points[i].y);
    if (p.getType() == polyPart.filledPolygon) PSFillPath();
    else {
      if (thickness >= 0) {
        PSStrokeThickPath(p.getType() == polyPart.polygon, thickness);
      }
      else PSStrokePath(p.getType() == polyPart.polygon);
    }
  }
  
  private void PSOutputBezierPart(bezierPart p, double thickness) {
    point[][] points = p.getPoints();
    double[] bounds = p.bounds();
    PSStartPath(bounds[0], bounds[1], bounds[2], bounds[3]);
    for (int i = 0; i < points.length-1; i++) {
      PSNextBezierSegment(points[i][1].x, points[i][1].y,
                        points[i][2].x, points[i][2].y,
                        points[i+1][0].x, points[i+1][0].y,
                        points[i+1][1].x, points[i+1][1].y);
    }
    if (p.getType() == bezierPart.filledCurve ||
       p.getType() == bezierPart.closedCurve) {
      PSNextBezierSegment(points[points.length-1][1].x, points[points.length-1][1].y,
                        points[points.length-1][2].x, points[points.length-1][2].y,
                        points[0][0].x, points[0][0].y,
                        points[0][1].x, points[0][1].y);
    }
    if (p.getType() == bezierPart.filledCurve) PSFillPath();
    else {
      if (thickness >= 0) {
        PSStrokeThickPath(p.getType() == bezierPart.closedCurve, thickness);
      }
      else PSStrokePath(p.getType() == bezierPart.closedCurve);
    }
  }
  
  private void PSOutputPreamble() {
    if (clip) {
      Dimension d = theCanvas.getSize();
      Point t = theCanvas.getTranslation();
      double s = theCanvas.getScaling();
      PSPreamble((-t.x - d.width / 2) / s, (t.y - d.height / 2) / s,
                 (-t.x + d.width / 2) / s, (t.y + d.height / 2) / s);
    }
    else {
      double[] bounds = displayed.bounds();
      PSPreamble(bounds[0], bounds[1], bounds[2], bounds[3]);
    }
  }
  
  private final class displayCanvas extends canvas {
  
    public void paint(Graphics g) {
      super.paint(g);
      int type = Cursor.WAIT_CURSOR;
      Cursor wait = Cursor.getPredefinedCursor(type);
      theCanvas.setCursor(wait);

      if (displayed != null) {
        Dimension size = getSize();
        if (autoZoom) {
          double[] bounds = displayed.bounds();
          bounds[2] -= bounds[0];
          bounds[3] -= bounds[1];
          setScaling(Math.min(((double)size.width)/bounds[2]*autoZoomSize,
                                        ((double)size.height)/bounds[3]*autoZoomSize));
          if (getScaling() == Double.POSITIVE_INFINITY) setScaling(1);
          setTranslation(-(int)((bounds[0] + bounds[2]/2) * getScaling()),
                                 (int)((bounds[1] + bounds[3]/2) * getScaling()));
        }
        g.translate(size.width/2 + getTranslation().x, size.height/2 + getTranslation().y);
        Enumeration parts = displayed.parts();
        while (parts.hasMoreElements()) {
          part p = (part)parts.nextElement();
          if (!(p.shown = autoZoom || insideWindow(p))) continue; // remember whether p is visible
          g.setColor(p.getColour());
          if (p instanceof dot) drawDot(g, (dot)p);
          else if (p instanceof polyPart) drawPolyPart(g, (polyPart)p);
          else if (p instanceof bezierPart) drawBezierPart(g, (bezierPart)p);
        }
      }
      int oldtype = Cursor.DEFAULT_CURSOR;
      Cursor normal = Cursor.getPredefinedCursor(oldtype);
      theCanvas.setCursor(normal);
    }
    
    private boolean insideWindow(part p) {
      Dimension size = getSize();
      size.width /= 2;
      size.height /= 2;
      double[] b = p.bounds();
      b[0] = b[0] * getScaling() + getTranslation().x;
      b[1] = b[1] * getScaling() - getTranslation().y;
      b[2] = b[2] * getScaling() + getTranslation().x;
      b[3] = b[3] * getScaling() - getTranslation().y;
      return !(b[0] < -size.width && b[2] < -size.width || b[0] > size.width && b[2] > size.width)
        && !(b[1] < -size.height && b[3] < -size.height || b[1] > size.height && b[3] > size.height);
    }
    
    private void drawDot(Graphics g, dot d) {
      point position = d.getPosition();
      double x = position.x * getScaling();
      double y = -position.y * getScaling();
      if (x <= Integer.MAX_VALUE && x >= Integer.MIN_VALUE &&
          y <= Integer.MAX_VALUE && y >= Integer.MIN_VALUE) {
        int size = (int)Math.ceil(d.getThickness() * getScaling());
        if (size <= 1) g.drawRect((int)Math.round(x), (int)Math.round(y), 0,0);
        else {
          g.drawOval((int)Math.round(x-size/2.0), (int)Math.round(y-size/2.0), size, size);
          g.fillOval((int)Math.round(x-size/2.0), (int)Math.round(y-size/2.0), size, size);
        }
      }
    }
    
    private void drawPolyPart(Graphics g, polyPart p) {
      point[] points;
      int[] xPoints, yPoints;
      points = p.getPoints();
      if (p.getType() == polyPart.polygon || p.getType() == polyPart.filledPolygon) {
        xPoints = new int[points.length+1];
        yPoints = new int[points.length+1];
      }
      else {
        xPoints = new int[points.length];
        yPoints = new int[points.length];
      }
      for (int i = 0; i < points.length; i++) {
        double x = points[i].x*getScaling();
        double y = -points[i].y*getScaling();
        if (x <= Integer.MAX_VALUE && x >= Integer.MIN_VALUE &&
            y <= Integer.MAX_VALUE && y >= Integer.MIN_VALUE) {
          xPoints[i] = (int)Math.round(x);
          yPoints[i] = (int)Math.round(y);
        }
       else return;
      }
      if (p.getType() == polyPart.polygon || p.getType() == polyPart.filledPolygon) {
        xPoints[points.length] = xPoints[0];
        yPoints[points.length] = yPoints[0];
      }
      if (p.getType() == polyPart.filledPolygon) {
        g.drawPolyline(xPoints, yPoints, xPoints.length);
        g.fillPolygon(xPoints, yPoints, xPoints.length);
      }
      else if (p.getThickness() < 0) g.drawPolyline(xPoints, yPoints, xPoints.length);
      else drawThickLines(g, xPoints, yPoints, p.getThickness() * getScaling());
    }
    
    private void drawBezierPart(Graphics g, bezierPart p) {
      point[] points;
      int[] xPoints, yPoints;
      points = p.toPolyline(1/getScaling());
      xPoints = new int[points.length];
      yPoints = new int[points.length];
      for (int i = 0; i < points.length; i++) {
        double x = points[i].x*getScaling();
        double y = -points[i].y*getScaling();
        if (x <= Integer.MAX_VALUE && x >= Integer.MIN_VALUE &&
            y <= Integer.MAX_VALUE && y >= Integer.MIN_VALUE) {
          xPoints[i] = (int)Math.round(x);
          yPoints[i] = (int)Math.round(y);
        }
       else return;
      }
      if (p.getType() == bezierPart.filledCurve) {
        g.drawPolyline(xPoints, yPoints, xPoints.length);
        g.fillPolygon(xPoints, yPoints, xPoints.length);
      }
      else if (p.getThickness() < 0) g.drawPolyline(xPoints, yPoints, xPoints.length);
      else drawThickLines(g, xPoints, yPoints, p.getThickness() * getScaling());
    }
    
    private void drawThickLines(Graphics g, int[] x, int[] y, double thickness) {
      int numSeg = x.length-1;
      int[][] a = new int[numSeg][4];
      int[][] b = new int[numSeg][4];
      for (int i = 0; i < numSeg; i++) {
        int deltaX = x[i+1] - x[i];
        int deltaY = y[i+1] - y[i];
        double angle = Math.atan2(-deltaX, deltaY);
        double a0, b0;
        a0 = Math.cos(angle) * thickness / 2;
        if (deltaY == 0) b0 = thickness / 2;
        else b0 = a0 * (-(double)deltaX / (double)deltaY);
        a[i][0] = (int)Math.round(a0 + x[i]);
        b[i][0] = (int)Math.round(b0 + y[i]);
        a[i][3] = (int)Math.round(x[i] - a0);
        b[i][3] = (int)Math.round(y[i] - b0);
        a[i][1] = a[i][0] + deltaX;
        b[i][1] = b[i][0] + deltaY;
        a[i][2] = a[i][3] + deltaX;
        b[i][2] = b[i][3] + deltaY;
      }
      numSeg--;
      for (int i = 0; i <= numSeg; i++) {
        g.fillPolygon(a[i], b[i], 4);
        if (i < numSeg) {
          a[i][0] = a[i+1][0];
          b[i][0] = b[i+1][0];
          g.fillPolygon(a[i], b[i], 3);
          a[i][0] = a[i+1][3];
          b[i][0] = b[i+1][3];
          g.fillPolygon(a[i], b[i], 3);
        }
      }
    }

  }

}
