package generators;

import java.io.*;
import java.util.*;
import parsers.*;
import terms.*;
import util.*;

/** An implementation of ETOL tree grammars.
  */
public class ETOLTreeGrammar extends termGenerator {

  regularTreeGrammar gram;
  tdTransducer trans;

//====================================================================
// First the 'interactive' part that deals with commands ...
//====================================================================

  private static String enumerate = "table enumeration";
  private static String advance = "advance";
  private static String single = "derive stepwise";
  private static String complete = "results only";
  private static String step = "derivation step";
  private static String random = "random tables";
  private static String refine = "refine";
  private static String back = "back";
  private static String reset = "reset";
  private static String newSeed = "new derivation";
  private static String[][] eCommands = {{ advance, reset }, {random}};
  private static String[][] rCommands = {{ refine, back, reset }, {enumerate}};
  private static String[][] sCommands = {{ step, back }, { complete }};
  private static String[][] cCommands = {{ single }};
  private static String[] tCommands = { newSeed };

  public list commands() {
    list result = super.commands();
    if (gram.isEnumerate) {
      for (int i = 0; i < eCommands.length; i++) result.append(eCommands[i]);
      if (gram.isStepwise) for (int i = sCommands.length; i > 0;) result.tail().prepend(sCommands[--i]);
      else for (int i = cCommands.length; i > 0;) result.tail().prepend(cCommands[--i]);
    }
    else  for (int i = 0; i < rCommands.length; i++) result.append(rCommands[i]);
    if (!trans.isDeterministic()) result.append(tCommands);
    return result;
  }
  
  public void execute(String com) {
    recompute = true;
    if (enumerate.equals(com)) gram.execute("enumeration");
    else if (advance.equals(com)) gram.execute("advance");
    else if (single.equals(com)) gram.execute("derive stepwise");
    else if (complete.equals(com)) gram.execute("results only");
    else if (step.equals(com)) gram.execute("single step");
    else if (random.equals(com)) gram.execute("random generation");
    else if (refine.equals(com)) gram.execute("refine");
    else if (back.equals(com)) gram.execute("back");
    else if (reset.equals(com)) gram.execute("reset");
    else if (newSeed.equals(com)) trans.execute("new random seed");
    else super.execute(com);
  }

//====================================================================
// Now the actual implementation of ETOL tree grammars ...
//====================================================================

  private static String grammarExtension = ".1";
  private static String transducerExtension = ".2";
  
  private finiteSignature outputSig;
  private finiteSignature tableSig = new finiteSignature();
  private fixedRankSignature nontermSig;
  private fixedRankSignature origNontermSig;
  private fixedRankSignature grammarNonterminals = new fixedRankSignature(0);
  private String startSym;
  private Vector tableNames;
  private Vector tables;
  private Vector regularRules = new Vector();
  private Vector chainRules = new Vector();
  private rexp regulation = null;
  private BitSet terminatingTable;
  private BitSet nonterminatingTable;
  private symbol bottom = new symbol("bot", 0);
  
  private term current = null;
  private boolean recompute = true;
  
  public term currentTerm() {
    if (recompute) {
      try { current = trans.apply(gram.currentTerm()); }
      catch (ExitException e) { return current = null; }
      recompute = false;
    }
    return current;
  }
  
/** Initialize this <code>ETOLTreeGrammar</code> by reading its definition from a stream.
  * The syntax is defined by the class <code>ETOLTreeGrammarParser</code>.
  * Before the grammar can be used, <code>translate</code> must be called with a
  * filename (which is used to store the regular tree grammar and the top-down tree
  * transducer into which the grammar is translated).
  * @see ETOLTreeGrammarParser
  * @exception ParseException if an error occurs
  */
  public void parse(ASCII_CharStream stream) throws ParseException {
    ETOLTreeGrammarParser parser = new ETOLTreeGrammarParser(stream);
    parser.ETOLTreeGrammar();
    outputSig = parser.sig;
    nontermSig = new fixedRankSignature(1);
    origNontermSig = parser.nonterminals;
    Enumeration enum = origNontermSig.elements();
    while (enum.hasMoreElements()) {
      nontermSig.addSymbol(((symbol)enum.nextElement()).toString());
    }
    startSym = parser.startSymbol;
    tableNames = parser.tableNames;
    tables = parser.tables;
    regulation = parser.regulation;
    addImplicitRules();
    determineTermination();
    computeTableSig();
  }
  
  private void addImplicitRules() {
    for (int t=0; t < tables.size(); t++) {
      Vector collect = new Vector();
      Vector table = (Vector)tables.elementAt(t);
      Enumeration enum = nontermSig.elements();
      search:
      while (enum.hasMoreElements()) {
        symbol s = (symbol)enum.nextElement();
        for (int i=0; i < table.size(); i++) {
          if (s.equals(lhsSymbol(table,i))) continue search;
        }
        collect.addElement(s);
      }
      String tableName = (String)tableNames.elementAt(t);
      for (int i=0; i < collect.size(); i++) {
        Object[] rule = new Object[3];
        table.addElement(rule);
        symbol sym = new variable(1);
        term aux = new term(sym);
        sym = new symbol(tableName, 1);
        term aux2 = new term(sym);
        aux2.defineSubterm(0, aux);
        aux = aux2;
        aux2 = new term((symbol)collect.elementAt(i));
        aux2.defineSubterm(0, aux);
        rule[0] = aux2; // left-hand side
        sym = new variable(1);
        aux = new term(sym);
        aux2 = new term((symbol)collect.elementAt(i));
        aux2.defineSubterm(0, aux);
        rule[1] = aux2; // right-hand side
        rule[2] = new Double(1); // weight
      }
    }
  }
  
  private void determineTermination() {
    terminatingTable = new BitSet(tables.size());
    nonterminatingTable = new BitSet(tables.size());
    for (int t=0; t < tables.size(); t++) {
      boolean termFailed = false;
      boolean nontermFound = false;
      Vector table = (Vector)tables.elementAt(t);
      Enumeration enum = nontermSig.elements();
      while (enum.hasMoreElements()) {
        symbol s = (symbol)enum.nextElement();
        boolean termFound = false;
        for (int i=0; i < table.size(); i++) {
          if (s.equals(lhsSymbol(table,i))) {
            if (isTerminal(rhsTerm(table,i))) termFound = true;
            if (isNonterminal(rhsTerm(table,i))) nontermFound = true;
          }
        }
        if (!termFound) termFailed = true;
      }
      if (!termFailed) terminatingTable.set(t);
      if (nontermFound) nonterminatingTable.set(t);
    }
  }
  
  private symbol lhsSymbol(Vector table, int ruleNo) {
    Object[] rule = (Object[])table.elementAt(ruleNo);
    term t = (term)rule[0];
    return t.topSymbol();
  }
  
  private term lhsTerm(Vector table, int ruleNo) {
    Object[] rule = (Object[])table.elementAt(ruleNo);
    return (term)rule[0];
  }
  
  private term rhsTerm(Vector table, int ruleNo) {
    Object[] rule = (Object[])table.elementAt(ruleNo);
    return (term)rule[1];
  }
  
  private double weight(Vector table, int ruleNo) {
    Object[] rule = (Object[])table.elementAt(ruleNo);
    return ((Double)rule[2]).doubleValue();
  }
  
  private boolean isTerminal(term t) {
    symbol top = t.topSymbol();
    if (nontermSig.contains(top)) return outputSig.contains(new symbol(top.toString(),0));
    for (int i=0; i < top.rank(); i++) {
      if (!isTerminal(t.subterm(i))) return false;
    }
    return true;
  }
  
  private boolean isNonterminal(term t) {
    symbol top = t.topSymbol();
    if (nontermSig.contains(top)) return true;
    for (int i=0; i < top.rank(); i++) {
      if (isNonterminal(t.subterm(i))) return true;
    }
    return false;
  }
  
  private void computeTableSig() {
    boolean terminalTableExists = false;
    for (int i=0; i < tables.size(); i++) {
      String tableName = (String)tableNames.elementAt(i);
      if (regulation != null) tableSig.addSymbol(new symbol(tableName,1));
      else {
        if (nonterminatingTable.get(i)) tableSig.addSymbol(new symbol(tableName,1));
        if (terminatingTable.get(i)) {
          tableSig.addSymbol(new symbol(tableName,0));
          terminalTableExists = true;
        }
      }
    }
    if (!terminalTableExists) tableSig.addSymbol(bottom);
  }
  
  public void translate(String fileName) throws ParseException {
    File file1 = new File(fileName + grammarExtension);
    File file2 = new File(fileName + transducerExtension);
    try {
      if (regulation == null) generateDefaultGrammar(openOutFile(file1));
      else generateGrammar(openOutFile(file1), regulation);
      generateTransducer(openOutFile(file2));
    }
    catch (IOException e) {
      throw new ParseException("Could not write to output file:\n" + e.getMessage());
    }
    try {
      objectParser parser = new objectParser(new ASCII_CharStream(new FileInputStream(file1),1,1));
      gram = (regularTreeGrammar)parser.parse();
      parser = new objectParser(new ASCII_CharStream(new FileInputStream(file2),1,1));
      trans = (tdTransducer)parser.parse();
    }
    catch (ParseException e) {
      throw new ParseException("Could not parse generated file:\n" +
        e.getMessage() +
        "\n[This should not happen. There must be one of those " +
        "small insects in the implementation 8-( ]");
    }
    catch (FileNotFoundException e) {
      throw new ParseException("Could not find generated file:\n" + e.getMessage() +
        "\n[This should not happen. There must be one of those " +
        "small insects in the implementation 8-( ]");
    }
  }
  
  private FileWriter openOutFile(File file) throws IOException {
    if (file.exists()) file.delete();
    FileWriter result = new FileWriter(file);
    return result;
  }
  
  private symbol grammarSymbol(int n) {
    return new symbol("S" + n, 0);
  }
  
  private void generateDefaultGrammar(FileWriter out) throws IOException {
    grammarNonterminals.addSymbol("S0");
    out.write("generators.regularTreeGrammar:\n");
    out.write("  (\n");
    out.write("    " + grammarNonterminals + ",\n");
    out.write("    " + tableSig + ",\n");
    out.write("    {\n");
    tableEnumerationRules(out);
    out.write("    },\n");
    out.write("    " + grammarSymbol(0) + "\n");
    out.write("  )");
    out.close();
  }
  
  private void tableEnumerationRules(FileWriter out) throws IOException {
    Enumeration enum = tableSig.elements();
    while (enum.hasMoreElements()) {
      symbol s = (symbol)enum.nextElement();
      if (s.rank() == 1) out.write("      " + grammarSymbol(0) + " -> " + s.toString() + "[" + grammarSymbol(0) + "]");
      else out.write("      " + grammarSymbol(0) + " -> " + s.toString() + " weight 0");
      if (enum.hasMoreElements()) out.write(",\n");
      else out.write("\n");
    }
  }
  
  private void generateGrammar(FileWriter out, rexp r) throws IOException {
    int max = generateRules(r, 0);
    for (int i = chainRules.size(); i <= max; i++) chainRules.addElement(new Vector());
    for (int i = regularRules.size(); i <= max; i++) regularRules.addElement(new Vector());
    addRegularRule(max, bottom + " weight 0");
    applyClosure();
    for (int i = 0; i <= max; i++) grammarNonterminals.addSymbol("S" + i);
    out.write("generators.regularTreeGrammar:\n");
    out.write("  (\n");
    out.write("    " + grammarNonterminals + ",\n");
    out.write("    " + tableSig + ",\n");
    out.write("    {\n");
    for (int i = 0; i < regularRules.size(); i++) {
      Vector v = (Vector)regularRules.elementAt(i);
      for (int j = 0; j < v.size(); j++) {
        out.write("      S" + i + " -> " + v.elementAt(j));
        if (i+1 < regularRules.size() || j+1 < v.size()) out.write(",\n"); else out.write("\n");
      }
    }
    out.write("    },\n");
    out.write("    " + grammarSymbol(0) + "\n");
    out.write("  )");
    out.close();
  }
  
  boolean loopFlag = false;
      
  private int generateRules(rexp r, int n) {
    switch (r.type) {
      case rexp.CONC:
        boolean oldLoopFlag = loopFlag;
        loopFlag = r.subexp[1].containsLoop();
        int m = generateRules(r.subexp[0], n);
        loopFlag = oldLoopFlag;
        return generateRules(r.subexp[1], m);
      case rexp.UNION:
        oldLoopFlag = loopFlag;
        int m1 = generateRules(r.subexp[0], n+1);
        boolean newLoopFlag = loopFlag;
        loopFlag = oldLoopFlag;
        int m2 = generateRules(r.subexp[1], m1 + 1);
        loopFlag = loopFlag || newLoopFlag;
        addChainRule(n, n+1);
        addChainRule(n, m1+1);
        addChainRule(m1, m2);
        return m2;
      case rexp.PLUS:
        loopFlag = true;
        m = generateRules(r.subexp[0], n);
        addChainRule(m, n);
        addChainRule(m, m + 1);
        return m+1;
      case rexp.STAR:
        loopFlag = true;
      case rexp.OPT:
        m = generateRules(r.subexp[0], n+1);
        addChainRule(n, m + 1);
        addChainRule(n, n + 1);
        if (r.type == rexp.STAR) addChainRule(m, n + 1);
        addChainRule(m, m + 1);
        return m+1;        
      default:
        if (loopFlag) addRegularRule(n, "t" + r.type + "[S" + (n+1) + "]");
        else addRegularRule(n, "t" + r.type + "[S" + (n+1) + "] weight 0");
        return n+1;
    }
  }
  
  private void addChainRule(int lhs, int rhs) {
    for (int i = chainRules.size(); i <= lhs; i++) chainRules.addElement(new Vector());
    ((Vector)chainRules.elementAt(lhs)).addElement(new Integer(rhs));
  }
  
  private void addRegularRule(int lhs, String rhs) {
    for (int i = regularRules.size(); i <= lhs; i++) regularRules.addElement(new Vector());
    ((Vector)regularRules.elementAt(lhs)).addElement(rhs);
  }
  
  private void applyClosure() {
    closure();
    Vector result = new Vector(regularRules.size());
    for (int i = 0; i < chainRules.size(); i++) {
      Vector v = new Vector();
      result.addElement(v);
      Vector cr = (Vector)chainRules.elementAt(i);
      for (int j = 0; j < cr.size(); j++) {
        int k = ((Integer)cr.elementAt(j)).intValue();
        v.addAll((Vector)regularRules.elementAt(k));
      }
    }
    regularRules = result;
  }
  
  private void closure() {
    for (int i = 0; i < chainRules.size(); i++) {
      Vector cr = (Vector)chainRules.elementAt(i);
      cr.addElement(new Integer(i));
    }
    boolean changed;
    do {
      changed = false;
      for (int i = 0; i < chainRules.size(); i++) {
        Vector cr = (Vector)chainRules.elementAt(i);
        for (int j = 0; j < cr.size(); j++) {
          int k = ((Integer)cr.elementAt(j)).intValue();
          Vector cr2 = (Vector)chainRules.elementAt(k);
          for (int l = 0; l < cr2.size(); l++) {
            if (!cr.contains(cr2.elementAt(l))) {
              cr.addElement(cr2.elementAt(l));
              changed = true;
            }
          }
        }
      }
    } while (changed);
  }
  
  private void generateTransducer(FileWriter out) throws IOException {
    out.write("generators.tdTransducer:\n");
    out.write("  (\n");
    finiteSignature extendedTableSig = new finiteSignature();
    extendedTableSig.unionWith(tableSig);
    extendedTableSig.unionWith(grammarNonterminals);
    out.write("    " + extendedTableSig + ",\n");
    finiteSignature extendedOutputSig = new finiteSignature();
    extendedOutputSig.unionWith(origNontermSig);
    extendedOutputSig.unionWith(outputSig);
    out.write("    " + extendedOutputSig.toString() + ",\n");
    out.write("    " + nontermSig.toString() + ",\n");
    out.write("    {\n");
    tableImplementationRules(out);
    out.write("    },\n");
    out.write("    " + startSym + "\n");
    out.write("  )");
    out.close();
  }
  
  private void tableImplementationRules(FileWriter out) throws IOException {
    Vector rules = new Vector();
    for (int t=0; t < tables.size(); t++) {
      Vector table = (Vector)tables.elementAt(t);
      if (regulation != null || nonterminatingTable.get(t)) {
        for (int i=0; i < table.size(); i++) {
          term left = lhsTerm(table, i);
          term right = rhsTerm(table, i);
          double weight = weight(table, i);
          if (weight == 1) rules.addElement(left.toString() + " -> " + right.toString());
          else rules.addElement(left.toString() + " -> " + right.toString() + " weight " + weight);
        }
      }
      if (regulation == null && terminatingTable.get(t)) {
        for (int i=0; i < table.size(); i++) {
          if (isTerminal(rhsTerm(table, i))) {
            term left = lhsTerm(table, i);
            term leftSub = left.subterm(0);
            leftSub.relabel(new symbol(leftSub.topSymbol().toString(), 0));
            term right = terminal(rhsTerm(table, i));
            double weight = weight(table, i);
            if (weight == 1) rules.addElement(left.toString() + " -> " + right.toString());
            else rules.addElement(left.toString() + " -> " + right.toString() + " weight " + weight);
          }
        }
      }
    }
    Enumeration enum1 = nontermSig.elements();
    while (enum1.hasMoreElements()) {
      symbol s1 = (symbol)enum1.nextElement();
      if (outputSig.contains(new symbol(s1.toString(),0)) && tableSig.contains(bottom)) {
        rules.addElement(s1 + "[" + bottom + "] -> " + s1);
      }
      Enumeration enum2 = grammarNonterminals.elements();
      while (enum2.hasMoreElements()) {
        symbol s2 = (symbol)enum2.nextElement();
        rules.addElement(s1 + "[" + s2 + "] -> " + s1);
      }
    }
    for (int i=0; i < rules.size(); i++) {
      out.write("      " + (String)rules.elementAt(i));
      if (i+1 < rules.size()) out.write(", \n");
      else out.write("\n");
    }
  }
  
  private term terminal(term t) {
    symbol top = t.topSymbol();
    if (nontermSig.contains(top)) return new term(new symbol(top.toString(),0));
    term result = new term(top);
    for (int i=0; i < top.rank(); i++) {
      result.defineSubterm(i, terminal(t.subterm(i)));
    }
    return result;
  }
  
}

