package generators;

import java.util.*;
import terms.*;
import parsers.*;
import util.*;

/** A top-down tree transducer.
  * Currently, non-determinism is implemented by random choice (evenly distributed, if
  * <code>Math.random()</code> yields such a distribution). This is not really the most
  * appropriate choice, but it is simple...<br>
  * The basic behaviour of an object of this class can be controlled by the commands
  * <code>single steps</code> and <code>complete runs</code>. Normally, if an input
  * term is given to the transducer, rules will be applied as long as possible and, upon
  * successful termination, the resulting term is returned (otherwise, the result is
  * <code>null</code>). However, using <code>single steps</code> one can switch to
  * a mode where single derivation steps are applied in a breadth-first manner. In this
  * case, there is an additional command <code>next step</code> which acts in the
  * obvious way. Using the command <code>complete runs</code> one can switch back
  * to the default behaviour. In case of non-determinism, the sequences of random choices
  * underlying successive derivations will be based on one and the same random seed
  * unless the command <code>new random seed</code> is applied, which randomly
  * selects a new seed. 
  */
public class tdTransducer extends termModifier {

//====================================================================
// First the 'interactive' part that deals with commands ...
//====================================================================
  private static String nextStep = "single step";
  private static String parallelStep = "parallel step";
  private static String back = "back";
  private static String singleSteps = "derive stepwise";
  private static String completeRuns = "results only";
  private static String newSeed = "new random seed";
  private static String variable = "variable random seed";
  private static String fixed = "fixed random seed";
  private boolean stepwise = false;
  protected boolean fixedSeed = true;
  private Vector stepsDone = new Vector();
  private boolean isDeterministic;
  
/** Returns the (names of the) commands described above. */
  public list commands() {
    list result = new list();
    String[] com;
    if (stepwise) {
      com = new String[3];
      com[0] = nextStep;
      com[1] = parallelStep;
      com[2] = back;
      result.append(com);
    }
    if (isDeterministic) {
      com = new String[1];
      com[0] = stepwise ? completeRuns : singleSteps;
      result.append(com);
    }
    else {
      com = new String[2];
      com[0] = stepwise ? completeRuns : singleSteps;
      com[1] = newSeed;
      result.append(com);
      com = new String[1];
      com[0] = fixedSeed ? variable : fixed;
      result.append(com);
    }
    return result;
  }
  
  public void execute(String command) {
    if (nextStep.equals(command)) {
      singleStep();
      stepsDone.addElement(nextStep);
    }
    else if (parallelStep.equals(command)) {
      parallelStep();
      stepsDone.addElement(parallelStep);
    }
    else if (back.equals(command) && stepsDone.size()>0) {
      stepsDone.removeElementAt(stepsDone.size()-1);
      run(lastArgument);
      for (int i = 0; i < stepsDone.size(); i++) {
        if (stepsDone.elementAt(i) == nextStep) singleStep();
        else parallelStep();
      }
    }
    else if (newSeed.equals(command)) {
      newSeed();
      run(lastArgument);
      if (stepwise) stepsDone.removeAllElements();
    }
    else {
      if (variable.equals(command) || fixed.equals(command)) fixedSeed = !fixedSeed;
      else if (singleSteps.equals(command)) {
        stepwise = true;
        resetSeed();
        startComputation(lastArgument);
      }
      else if (completeRuns.equals(command)) {
        stepwise = false;
        stepsDone.removeAllElements();
        run(lastArgument);
      }
    }
  }

//====================================================================
// Now the actual implementation of top-down tree transducers ...
//====================================================================

  protected term[][] rule;
  protected double[] weight;
  protected signature states;
  protected signature inputSignature;
  protected symbol initialState;
  private term lastArgument = null;
  private term currTerm = null;
  protected list redices;
  private long randomSeed;
  private Random random = new Random();
  
  public tdTransducer() {
    redices = new list();
    newSeed();
  }
  
  protected void newSeed() {
    randomSeed = (long)(((double)Long.MAX_VALUE) * Math.random());
    random.setSeed(randomSeed);
  }
  
  protected void resetSeed() {
    if (fixedSeed) random.setSeed(randomSeed);
    else newSeed();
  }
  
  private boolean startComputation(term t) {
    if (validInput(t)){
      currTerm = new term(initialState);
      currTerm.defineSubterm(0, makeWorkingCopy(t));
      redices = new list();
      redices.append(currTerm);
      return true;
    }
    else {
      currTerm = null;
      return false;
    }
  }

/** Compute the output for a given input term. */
  public term apply(term t) throws ExitException {
    lastArgument = t;
    resetSeed();
    stepsDone.removeAllElements();
    run(t);
    return currTerm;
  }
  
  protected term makeWorkingCopy(term t) {
    return (term)t.clone();
  }
  
  private void run(term t)  {
    if (startComputation(t)){
      if (!stepwise) {
        while (!redices.isEmpty()) {
          parallelStep();
        }
      }
    }
    else currTerm = null;
  }
  
  private void singleStep() {
    if (currTerm != null && !redices.isEmpty()) {
      term result = oneStep();
      if (result == null) {
        redices = new list();
        currTerm = null;
      }
      else {
        list newRed = newRedices(result);
        newRed.concat(redices);
        redices = newRed;
      }
    }
  }
  
  private void parallelStep() {
    allowExit();
    if (currTerm != null) {
      while (!redices.isEmpty()) {
        if (oneStep() == null) {
          redices = new list();
          currTerm = null;
          return;
        }
      }
      redices.concat(newRedices(currTerm));
    }
  }
  
  private boolean validInput(term t) {
    if (t != null && inputSignature.contains(t.topSymbol())) {
      boolean result = true;
      for (int i = 0; i < t.topSymbol().rank(); i++) result &= validInput(t.subterm(i));
      return result;
    }
    else return false;
  }
  
  protected term oneStep() {
    term first = (term)redices.head();
    term[] matching = null;
    double sumWeights = 0;
    int ruleNumber = 0;
    for (int i = 0; i < rule.length; i++) {
      term[] thisMatching =  rule[i][0].match(first);
      if (thisMatching != null && weight[i] > 0) {
        sumWeights += weight[i];
        if (random.nextFloat() <= weight[i]/sumWeights) {
          matching = thisMatching;
          ruleNumber = i;
        }
      }
    }
    if (sumWeights == 0) return null;
    term newFirst = rule[ruleNumber][1].substitute(matching);
    symbol top = newFirst.topSymbol();
    first.relabel(top);
    for (int i = 0; i < top.rank(); i++) first.defineSubterm(i, newFirst.subterm(i));
    redices = redices.tail();
    return first;
  }
  
  private list newRedices(term t) {
    symbol top = t.topSymbol();
    list result = new list();
    if (states.contains(top)) result.append(t);
    else {
      for (int i = 0; i < top.rank(); i++) result.concat(newRedices(t.subterm(i)));
    }
    return result;
  }
  
  public term currentTerm() {
    return currTerm;
  }
  
  public boolean isDeterministic() { return isDeterministic; }
  
  private void checkDeterminism() {
    isDeterministic = false;
    for (int i = 0; i < rule.length; i++) {
      for (int j = i+1; j < rule.length; j++) {
        if (rule[i][0].topSymbol().equals(rule[j][0].topSymbol()) &&
            rule[i][0].subterm(0).topSymbol().equals(rule[j][0].subterm(0).topSymbol())) {
          return;
        }
      }
    }
    isDeterministic = true;
  }

/** Initialize this object by parsing the definition of a top-down tree transducer.
  * The syntax is defined by the class <code>tdTransducerParser</code>.
  * @see tdTransducerParser
  * @exception ParseException if an error occurs
  */
  public void parse(ASCII_CharStream stream) throws ParseException {
    tdTransducerParser parser = new tdTransducerParser(stream);
    parser.tdTransducer();
    initialState = parser.initial;
    rule = parser.rule;
    weight = parser.weight;
    states = parser.states;
    inputSignature = parser.in;
    checkDeterminism();
  }
  
}
