package gui;

import java.awt.*;
import java.awt.event.*;

/** A canvas that allows the user to drag the shown image to another place using the mouse.
  * The <code>paint</code> method of subclasses of this class should take into account the
  * intended translation of the image that can be inquired using <code>getTranslation</code>,
  * in order to implement the expected behaviour.
  */
public class canvas extends Canvas {

  private Point translation = new Point(0,0);
  private double scaling = 1;
  private boolean scalingEnabled = true;
  private Point lastPressed = null;
  private Point firstPressed = null;
  
  public canvas() {
    super();
    enableEvents(AWTEvent.MOUSE_EVENT_MASK);
    enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);
  }

/** This yields the current translation of the picture.
  * It is affected by pressing the first mouse button and dragging the mouse
  * inside the region of the canvas.
  */
  public Point getTranslation() { return translation; }
  
  public void setTranslation(int x, int y) { translation.x = x; translation.y = y; }

/** This yields the current scaling factor.
  * It is affected by pressing the second mouse button and dragging the mouse
  * inside the region of the canvas.
  */
  public double getScaling() { return scaling; }
  
  public void setScaling(double newScaling) {
    double factor = newScaling / scaling;
    translation.x = (int)Math.round(translation.x * factor);
    translation.y = (int)Math.round(translation.y * factor);
    scaling = newScaling;
  }
  
  public void zoom(double factor) {
    translation.x = (int)Math.round(translation.x * factor);
    translation.y = (int)Math.round(translation.y * factor);
    scaling *= factor;
  }
  
  public void disableScaling() {
    scalingEnabled = false;
  }
  
  public void enableScaling() {
    scalingEnabled = true;
  }
  
  protected void processMouseEvent(MouseEvent e){
    super.processMouseEvent(e);
    if (e.getID() == MouseEvent.MOUSE_PRESSED) {
      if (button1(e)) lastPressed = e.getPoint();
      else if (button2(e) && scalingEnabled) {
        firstPressed = lastPressed = e.getPoint();
        drawSquare(getGraphics());
      }
    }
    else if (e.getID() == MouseEvent.MOUSE_RELEASED && lastPressed != null) {
      if (button1(e)) lastPressed = null;
      else if (button2(e) && scalingEnabled && firstPressed != null) {
        drawSquare(getGraphics());
        Dimension d = getSize();
        double newWidth = Math.abs(firstPressed.x - lastPressed.x);
        double newHeight = Math.abs(firstPressed.y - lastPressed.y);
        double factor = Math.min((double)d.width / newWidth, (double)d.height / newHeight);
        translation.x += ((int)(d.width - newWidth) / 2 - Math.min(firstPressed.x, lastPressed.x));
        translation.x = (int)Math.round(translation.x * factor);
        translation.y += ((int)(d.height - newHeight) / 2 - Math.min(firstPressed.y, lastPressed.y));
        translation.y = (int)Math.round(translation.y * factor);
        scaling *= factor;
        firstPressed = lastPressed = null;
        invalidate();
        repaint();
      }
    }
  }
  
  protected void processMouseMotionEvent(MouseEvent e){
    super.processMouseMotionEvent(e);
    if (e.getID() == MouseEvent.MOUSE_DRAGGED && lastPressed != null) {
      if (button1(e)) {
        Point p = e.getPoint();
        translation.x += p.x - lastPressed.x;
        translation.y += p.y - lastPressed.y;
        lastPressed = p;
        invalidate();
        repaint();
      }
      else if (button2(e) && scalingEnabled && firstPressed != null) {
        Graphics g = getGraphics();
        drawSquare(g);
        lastPressed = e.getPoint();
        drawSquare(g);
      }
    }
  }
  
  private void drawSquare(Graphics g) {
    g.setColor(Color.black);
    g.setXORMode(Color.white);
    g.drawRect((int)Math.min(firstPressed.x, lastPressed.x),
                    (int)Math.min(firstPressed.y, lastPressed.y),
                    (int)Math.abs(lastPressed.x - firstPressed.x),
                    (int)Math.abs(lastPressed.y - firstPressed.y));
  }
  
  private boolean button1(MouseEvent e) {
    return (e.getModifiers() == 0) ||  // This should not be possible, but it seems that
                                                       // there is a bug in some MouseEvent implementations
                                                       // causing the modifiers to be set to 0 if button 1 is
                                                       // pressed (maybe this depends on the window
                                                       // manager - I encountered it using fvwm2 on
                                                       // Solaris.
              ((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0 &&
              (e.getModifiers() & InputEvent.BUTTON2_MASK) == 0);
  }

  private boolean button2(MouseEvent e) {
    return (e.getModifiers() & InputEvent.BUTTON2_MASK) != 0 &&
           (e.getModifiers() & InputEvent.BUTTON1_MASK) == 0;
  }

}
