package gui.nodeTypes;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import gui.*;
import parsers.*;
import terms.*;
import util.*;

public class worksheetNode extends visibleNode implements Runnable, ActionListener {

  public parsable contents;
  public String fileName = null;
  public static String reload = "reload file";
  public static String delete = "delete";
  
  private static String[] commands = { reload, delete };
  
  public static int USER_COMMAND = 0;
  public static int TERM = 1;
  public static int STOP = 2;

  private commandPanel panel;
  private popupMenu menu;
  private exitableThread contentsThread;
  protected int commandType = -1;
  protected Object pendingCommand = null;
  private boolean inputPending = false;
  protected observable obs = new observable();
  
  public worksheetNode(parsable obj, String fileName) {
    super();
    this.fileName = fileName;
    contents = obj;
    addLabel(getName());
    addLabel("[class " + obj.getClassName() + "]");
    panel = new commandPanel(this);
    menu = new popupMenu(this);
    add(menu);
  }
  
  public void start() {
    updateCommandList();
    contentsThread = new exitableThread(this);
    contentsThread.start();
  }

  public String getName() { return contents.getName(); }
  
  public void initialize() throws ParseException {
  }
  
  public int openWindows() {
    return panel.isShowing() ? 1 : 0;
  }
  
  public boolean acceptsSource(worksheetNode source) {
    return !parent().existsPath(this, source);
  }
  
  public void mousePressed(MouseEvent e) {
    if (e.isPopupTrigger()) menu.show(this, e.getPoint().x, e.getPoint().y);
    else super.mousePressed(e);
  }
    
  public void actionPerformed(ActionEvent event) {
    componentGraph p = parent();
    p.addToActiveCount(1);
    String command = event.getActionCommand();
    if (reload.equals(command)) reload();
    else if (delete.equals(command)) delete();
    else transferCommand(command, USER_COMMAND);
    p.addToActiveCount(-1);
  }
  
  private void reload() {
    exitableThread oldContentsThread = contentsThread;
    transferCommand(null, STOP);
    try { oldContentsThread.join(); } catch (InterruptedException e) {}
    boolean ok = false;
    parsable parsed = null;
    FileInputStream input = null;
    try {
      input = new FileInputStream(fileName);
      objectParser parser = new objectParser(new ASCII_CharStream(input,1,1));
      parsed = parser.parse();
      ok = true;
    } catch (FileNotFoundException e) {
      errorDisplay.show("Parsing of `" + fileName + "' failed: file not found");
    } catch (IOException e) { // File could not be closed. I don't care.
    } catch (ParseException e) {
      errorDisplay.show("Parsing of `" + fileName + "' failed:\n" + e.getMessage());
    } catch (TokenMgrError e) {
      errorDisplay.show("Parsing of `" + fileName + "' failed:\n" + e.getMessage());
    }
    finally { if (input != null) try { input.close(); } catch (IOException e) {}}
    if (ok) {
      if (parsed.getClass().equals(contents.getClass())) {
        if (parsed.getName() == null) {
          parsed.setName("file " + fileName.substring(fileName.lastIndexOf(File.separatorChar)+1));
        }
        reinit(parsed);
      }
      else errorDisplay.show("Reload failed! Unable to reload '" + fileName +
                                           "' because the class of the component has been changed");
    }
    updateCommandList();
    contentsThread = new exitableThread(this);
    contentsThread.start();
    ping();
  }
  
  public void delete() {
    exitableThread oldContentsThread = contentsThread;
    transferCommand(null, STOP);
    try { oldContentsThread.join(); } catch (InterruptedException e) {}
    dispose();
    parent().deleteNode(this);
  }
  
  public synchronized void transferCommand(Object com, int type) {
    while (contentsThread != null && inputPending) {
      try { wait(); } catch (InterruptedException e) {}
    }
    if (contentsThread == null) return;
    commandType = type;
    pendingCommand = com;
    inputPending = true;
    contentsThread.exitRequest(true);
    notify();
  }
  
  protected void broadcastResult(term t) {
    obs.setChanged();
    obs.notifyObservers(t);
  }
  
  protected void open() {
    if (!panel.isShowing()) {
      Point p = parent().getLocationOnScreen();
      panel.setLocation(new Point(p.x + position.x + radius, p.y + position.y + radius));
    }
    panel.setVisible(true);
  }
  
  private void updateCommandList() {
    list com = (contents instanceof reactive) ? ((reactive)contents).commands() : new list();
    com.append(commands);
    panel.setCommands(com);
    menu.setCommands(com);
  }
  
  public void run() {
    Object command;
    int type;
    while (true) {
      synchronized (this) {
        while (!inputPending) {
          try { wait(); } catch (InterruptedException e) {}
        }
        command = pendingCommand;
        type = commandType;
        inputPending = false;
        contentsThread.exitRequest(false);
        notify();
      }
      if (type == STOP) {
        contentsThread = null;
        synchronized (this) { notifyAll(); }
        break;
      }
      else if (contents instanceof reactive) {
        try {
          parent().addToActiveCount(1);
          runContents(command, type);
        }
        catch (ExitException e) {}
        finally { parent().addToActiveCount(-1); }
      }
      updateCommandList();
    }
  }
  
  public void runContents(Object command, int type) {
    if (type == USER_COMMAND) ((reactive)contents).execute((String)command);
  }
  
  protected void dispose() {
    panel.dispose();
    menu.removeActionListener(this);
    parent().removeComponentListener(this);
    contents.dispose();
    obs.deleteObservers();
  }
  
  public void addObservingEdge(edge e) {
    obs.addObserver(e);
  }
  
  public void deleteObservingEdge(edge e) {
    obs.deleteObserver(e);
  }
  
  protected void reinit(parsable newContents) {
    parsable oldContents = contents;
    contents = newContents;
    try { initialize(); }
    catch (ParseException e) {
      errorDisplay.show("Initialization failed:\n" + e.getMessage());
      contents = oldContents;
      return;
    }
    oldContents.dispose();
    contents.setName(contents.getBasename());
    removeLabels();
    addLabel(contents.getName());
    addLabel("[class " + newContents.getClassName() + "]");
    invalidate();
    repaint();
    panel.setTitle(getName());
  }
  
  public void ping() {
  }
  
}
